@@ -107,6 +107,61 @@ pub fn map_memory(name: &str, size: target_ulong, addr: target_ptr_t) -> Result<
     }
 }
 
+const IS_32_BIT: bool = std::mem::size_of::<target_ptr_t>() == 4;
+const TARGET_BITS: usize = std::mem::size_of::<target_ptr_t>() * 8;
+
+fn hex_addr(addr: target_ptr_t) -> impl std::fmt::Display {
+    if IS_32_BIT {
+        format!("{:08x}", addr)
+    } else {
+        format!("{:016x}", addr)
+    }
+}
+
+pub fn virt_memory_dump(cpu: &mut CPUState, addr: target_ptr_t, len: usize) {
+    let memory = virtual_memory_read(cpu, addr, len).unwrap();
+
+    let start_addr_aligned = addr & !0xf;
+    let end_addr_aligned = (addr + (len as target_ptr_t)) & !0xf;
+
+    let bytes_offset = addr - start_addr_aligned;
+
+    let hex_dump = (start_addr_aligned..=end_addr_aligned)
+        .step_by(0x10)
+        .enumerate()
+        .map(|(line_num, line_addr)| {
+            let hex_data = (0..0x10)
+                .map(|offset_in_line| {
+                    if line_num == 0 && offset_in_line < (bytes_offset as usize) {
+                        "  ".into()
+                    } else {
+                        let byte_index = ((line_num * 0x10) + offset_in_line) - (bytes_offset as usize);
+
+                        if let Some(byte) = memory.get(byte_index) {
+                            format!("{:02x}", byte).into()
+                        } else {
+                            "  ".into()
+                        }
+                    }
+                })
+                .collect::<Vec<std::borrow::Cow<'static, str>>>()
+                .join(" ");
+
+            format!("{}║{}\n", hex_addr(line_addr), hex_data)
+        })
+        .collect::<String>();
+
+    println!(
+        "{} 00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f",
+        " ".repeat(TARGET_BITS / 4)
+    );
+    println!(
+        "{}╦═══════════════════════════════════════════════",
+        "═".repeat(TARGET_BITS / 4)
+    );
+    println!("{}", hex_dump);
+}
+
 // Private API ---------------------------------------------------------------------------------------------------------
 
 // https://stackoverflow.com/questions/59707349/cast-vector-of-i8-to-vector-of-u8-in-rust/59707887#59707887
